/*****************************************************************************
 * HASM64                                                                    *
 *****************************************************************************/

/*****************************************************************************
 * File name:     parser.c                                                   *
 * Created:       1997-09-04 by Hampa Hug <hhug@student.ethz.ch>             *
 * Last modified: 2002-08-27 by Hampa Hug <hhug@student.ethz.ch>             *
 * Copyright:     (C) 1997-2002 by Hampa Hug <hhug@student.ethz.ch>          *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by  the Free Software Foundation.                                         *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/

/* $Id: parser.c 10 2004-01-02 00:48:42Z hampa $ */


#include "hasm64.h"


int par_is_mnemonic (parser_t *par);
int par_mnemonic (parser_t *par);


parser_t *par_new (void)
{
  parser_t *par;

  par = (parser_t *) malloc (sizeof (parser_t));
  if (par == NULL) {
    return (NULL);
  }

  par->scn = NULL;
  par->st = NULL;
  par->cg = NULL;
  par->pass = 0;
  par->newpass = 0;
  par->label[0] = 0;
  par->mod_name = NULL;
  par->mod_base = 0;

  par->st = st_new ();
  if (par->st == NULL) {
    par_del (par);
    return (NULL);
  }

  par->cg = cg_new (256);
  if (par->cg == NULL) {
    par_del (par);
  }

  return (par);
}

void par_del (parser_t *par)
{
  if (par != NULL) {
    if (par->scn != NULL) {
      par->scn->del (par->scn);
    }

    if (par->st != NULL) {
      st_del (par->st);
    }

    if (par->cg != NULL) {
      cg_del (par->cg);
    }

    if (par->mod_name != NULL) {
      free (par->mod_name);
    }

    free (par);
  }
}

void par_set_base (parser_t *par, unsigned addr)
{
  par->mod_base = addr;
}

unsigned par_get_base (parser_t *par)
{
  return (par->mod_base);
}

void par_set_label (parser_t *par, char *str)
{
  str_copy (par->label, str, PAR_MAX_LABEL);
}

char *par_get_label (parser_t *par)
{
  return (par->label);
}

char *par_str (parser_t *par)
{
  return (par->scn->tok_str);
}

unsigned par_id (parser_t *par)
{
  return (par->scn->tok_id);
}

void par_set_scn (parser_t *par, scanner_t *scn)
{
  if (par->scn != NULL) {
    par->scn->del (par->scn);
  }

  par->scn = scn;
}

void par_scan (parser_t *par)
{
  scanner_t *scn;

  par->scn->scan (par->scn);

  while (par->scn->tok_id == SCN_TOK_NONE) {
    scn = scn64_get_next (par->scn);
    if (scn == NULL) {
      return;
    }

    scn64_set_next (par->scn, NULL);
    par->scn->del (par->scn);
    par->scn = scn;
  }
}

int par_op_db (parser_t *par)
{
  unsigned i, j, n;
  unsigned val;
  int      cont;
  char     *str;

  par_scan (par);

  if (par_id (par) == TOK_SYM_PERIOD) {
    par_scan (par);
    if (par_exp_val (par, &n)) {
      return (1);
    }
  }
  else {
    n = 1;
  }

  cont = 1;
  while (cont) {
    if (par_id (par) == TOK_STRING) {
      str = par_str (par);

      i = 0;
      while (str[i] != 0) {
        val = str[i];

        for (j = 0; j < n; j++) {
          cg_add_byte (par->cg, val & 0xff, 1);
          val >>= 8;
        }

        i += 1;
      }

      par_scan (par);
    }
    else {
      if (par_exp (par, &val)) {
        return (1);
      }

      for (i = 0; i < n; i++) {
        cg_add_byte (par->cg, val & 0xff, 1);
        val >>= 8;
      }
    }

    if (par_id (par) == TOK_SYM_COMMA) {
      par_scan (par);
    }
    else {
      cont = 0;
    }
  }

  return (0);
}

int par_ea (parser_t *par, unsigned *mode, unsigned *val)
{
  sumxy_t *sum;
  sumxy_t *tmp;

  sum = par_exp_xy (par);
  if (sum == NULL) {
    return (1);
  }

  if (sumxy_check (sum)) {
    err_print (par, ERR_ANY, "Illegal addressing mode");
    sumxy_del (sum);
    return (1);
  }

  if (sum->ind == NULL) {
    *mode = EA_IMM;
    *val = sum->val;
  }
  else {
    tmp = sum->ind;

    if (tmp->ind != NULL) {
      if ((tmp->val != 0) || (tmp->x != 0)) {
        sumxy_del (sum);
        return (1);
      }

      if (tmp->y != 0) {
        tmp = tmp->ind;

        if ((tmp->ind != NULL) || (tmp->x != 0) || (tmp->y != 0)) {
          sumxy_del (sum);
          return (1);
        }

        *mode = EA_IND_Y;
        *val = tmp->val;
      }
      else {
        tmp = tmp->ind;

        if ((tmp->ind != NULL) || (tmp->x != 1) || (tmp->y != 0)) {
          sumxy_del (sum);
          return (1);
        }

        *mode = EA_IND_X;
        *val = tmp->val;
      }
    }
    else {
      *val = tmp->val;

      if (tmp->x != 0) {
        if (tmp->val < 256) {
          *mode = EA_ZPG_X;
        }
        else {
          *mode = EA_ABS_X;
        }
      }
      else if (tmp->y != 0) {
        *mode = EA_ABS_Y;
      }
      else {
        if (tmp->val < 256) {
          *mode = EA_ZPG;
        }
        else {
          *mode = EA_ABS;
        }
      }
    }
  }

  sumxy_del (sum);

  return (0);
}

int par_op_ea (parser_t *par, unsigned op, unsigned modes)
{
  unsigned mode, val;

  par_scan (par);

  if (par_id (par) == TOK_SYM_A) {
    if ((modes & M_A) == 0) {
      err_print (par, ERR_ANY, "Illegal addressing mode");
      return (1);
    }

    cg_add_byte (par->cg, op + (EA_IMM << 2), 1);

    par_scan (par);

    return (0);
  }

  if (par_ea (par, &mode, &val)) {
    return (1);
  }

  if (((1 << mode) & modes) == 0) {
    err_print (par, ERR_ANY, "Illegal addressing mode");
    return (1);
  }

  cg_add_byte (par->cg, op + (mode << 2), 1);

  switch (mode) {
    case EA_ABS:
    case EA_ABS_X:
    case EA_ABS_Y:
      cg_add_byte (par->cg, val & 0xff, 1);
      cg_add_byte (par->cg, (val >> 8) & 0xff, 1);
      break;

    default:
      cg_add_byte (par->cg, val & 0xff, 1);
      break;
  }

  return (0);
}

int par_op_imp (parser_t *par, unsigned op)
{
  par_scan (par);

  cg_add_byte (par->cg, op, 1);

  return (0);
}

int par_op_bra (parser_t *par, unsigned cond)
{
  unsigned src, dst;

  par_scan (par);

  if (par_exp (par, &dst)) {
    return (1);
  }

  src = (cg_get_addr (par->cg) + 2) & 0xffff;

  if (par->newpass == 0) {
    if (src < dst) {
      if ((dst - src) > 127) {
        err_print (par, ERR_ANY, "Branch out of range");
        return (1);
      }
    }
    else {
      if ((src - dst) > 128) {
        err_print (par, ERR_ANY, "Branch out of range");
        return (1);
      }
    }
  }

  cg_add_byte (par->cg, 0x10 + ((cond & 7) << 5), 1);
  cg_add_byte (par->cg, (dst - src) & 0xff, 1);

  return (0);
}

int par_op_cpx (parser_t *par, unsigned op)
{
  unsigned mode, val;

  par_scan (par);

  if (par_ea (par, &mode, &val)) {
    return (1);
  }

  switch (mode) {
    case EA_IMM:
      cg_add_byte (par->cg, op, 1);
      cg_add_byte (par->cg, val & 0xff, 1);
      return (0);

    case EA_ZPG:
      cg_add_byte (par->cg, op + 4, 1);
      cg_add_byte (par->cg, val & 0xff, 1);
      return (0);

    case EA_ABS:
      cg_add_byte (par->cg, op + 0x0c, 1);
      cg_add_byte (par->cg, val & 0xff, 1);
      cg_add_byte (par->cg, (val >> 8) & 0xff, 1);
      return (0);

    default:
      err_print (par, ERR_ANY, "Illegal addressing mode");
      return (1);
  }

  return (1);
}

int par_op_ldxy (parser_t *par, unsigned op, int x)
{
  unsigned mode, val;

  par_scan (par);

  if (par_ea (par, &mode, &val)) {
    return (1);
  }

  switch (mode) {
    case EA_IMM:
      cg_add_byte (par->cg, op, 1);
      cg_add_byte (par->cg, val & 0xff, 1);
      return (0);

    case EA_ZPG:
      cg_add_byte (par->cg, op + 4, 1);
      cg_add_byte (par->cg, val & 0xff, 1);
      return (0);

    case EA_ABS:
      cg_add_byte (par->cg, op + 0x0c, 1);
      cg_add_byte (par->cg, val & 0xff, 1);
      cg_add_byte (par->cg, (val >> 8) & 0xff, 1);
      return (0);

    case EA_ABS_X:
    case EA_ZPG_X:
      if (x) {
        err_print (par, ERR_ANY,
          "Illegal adressing mode (ldx with [xxxx + x])"
        );
        return (1);
      }

      if (val < 256) {
        cg_add_byte (par->cg, op + 0x14, 1);
        cg_add_byte (par->cg, val & 0xff, 1);
      }
      else {
        cg_add_byte (par->cg, op + 0x1c, 1);
        cg_add_byte (par->cg, val & 0xff, 1);
        cg_add_byte (par->cg, (val >> 8) & 0xff, 1);
      }
      return (0);

    case EA_ABS_Y:
      if (x == 0) {
        err_print (par, ERR_ANY,
          "Illegal adressing mode (ldy with [xxxx + y])"
        );
        return (1);
      }

      if (val < 256) {
        cg_add_byte (par->cg, op + 0x14, 1);
        cg_add_byte (par->cg, val & 0xff, 1);
      }
      else {
        cg_add_byte (par->cg, op + 0x1c, 1);
        cg_add_byte (par->cg, val & 0xff, 1);
        cg_add_byte (par->cg, (val >> 8) & 0xff, 1);
      }
      return (0);

    default:
      err_print (par, ERR_ANY, "Illegal addressing mode");
      return (1);
  }

  return (1);
}

int par_op_stxy (parser_t *par, unsigned op, int x)
{
  unsigned mode, val;

  par_scan (par);

  if (par_ea (par, &mode, &val)) {
    return (1);
  }

  switch (mode) {
    case EA_ZPG:
      cg_add_byte (par->cg, op + 4, 1);
      cg_add_byte (par->cg, val & 0xff, 1);
      return (0);

    case EA_ABS:
      cg_add_byte (par->cg, op + 0x0c, 1);
      cg_add_byte (par->cg, val & 0xff, 1);
      cg_add_byte (par->cg, (val >> 8) & 0xff, 1);
      return (0);

    case EA_ZPG_X:
      if (x) {
        err_print (par, ERR_ANY,
          "Illegal addressing mode (stx with [xxxx + x])"
        );
        return (1);
      }

      if (val >= 256) {
        err_print (par, ERR_ANY,
          "Illegal addressing mode"
        );
        return (1);
      }

      cg_add_byte (par->cg, op + 0x14, 1);
      cg_add_byte (par->cg, val & 0xff, 1);
      return (0);

    case EA_ABS_Y:
      if (x == 0) {
        err_print (par, ERR_ANY,
          "Illegal adressing mode (sty with [xxxx + y])"
        );
        return (1);
      }

      if (val >= 256) {
        err_print (par, ERR_ANY,
          "Illegal addressing mode"
        );
        return (1);
      }

      cg_add_byte (par->cg, op + 0x14, 1);
      cg_add_byte (par->cg, val & 0xff, 1);
      return (0);

    default:
      err_print (par, ERR_ANY, "Illegal addressing mode");
      return (1);
  }

  return (1);
}

int par_op_jmp (parser_t *par)
{
  unsigned mode, val;

  par_scan (par);

  if (par_ea (par, &mode, &val)) {
    return (1);
  }

  switch (mode) {
    case EA_IMM:
      cg_add_byte (par->cg, 0x4c, 1);
      cg_add_byte (par->cg, val & 0xff, 1);
      cg_add_byte (par->cg, (val >> 8) & 0xff, 1);
      return (0);

    case EA_ABS:
    case EA_ZPG:
      cg_add_byte (par->cg, 0x6c, 1);
      cg_add_byte (par->cg, val & 0xff, 1);
      cg_add_byte (par->cg, (val >> 8) & 0xff, 1);
      return (0);

    default:
      err_print (par, ERR_ANY, "Illegal addressing mode");
      return (1);
  }

  return (1);
}

int par_op_jsr (parser_t *par)
{
  unsigned mode, val;

  par_scan (par);

  if (par_ea (par, &mode, &val)) {
    return (1);
  }

  switch (mode) {
    case EA_IMM:
      cg_add_byte (par->cg, 0x20, 1);
      cg_add_byte (par->cg, val & 0xff, 1);
      cg_add_byte (par->cg, (val >> 8) & 0xff, 1);
      return (0);

    default:
      err_print (par, ERR_ANY, "Illegal addressing mode");
      return (1);
  }

  return (1);
}

int par_op_rep (parser_t *par)
{
  unsigned val;
  unsigned i1, i2;

  i1 = cg_get_index (par->cg);

  par_scan (par);

  if (par_exp (par, &val)) {
    return (1);
  }

  if (val == 0) {
    err_print (par, ERR_ANY, "Repeat count zero");
    return (1);
  }

  if (!par_is_mnemonic (par)) {
    err_print (par, ERR_ANY, "Instruction expected");
    return (1);
  }

  if (par_mnemonic (par)) {
    return (1);
  }

  i2 = cg_get_index (par->cg) - 1;

  if ((val > 1) && (i1 <= i2)) {
    if (cg_rep_bytes (par->cg, i1, i2, val - 1)) {
      err_print (par, ERR_ANY, "Output exceeds 64K");
      return (1);
    }
  }

  return (0);
}

int par_is_mnemonic (parser_t *par)
{
  if (par->scn->tok_id < TOK_MNEMONIC_START) {
    return (0);
  }

  if (par->scn->tok_id > TOK_MNEMONIC_END) {
    return (0);
  }

  return (1);
}

int par_mnemonic (parser_t *par)
{
  switch (par_id (par)) {
    case TOK_OP_ADC:
      return (par_op_ea (par, 0x61, M_ALL));

    case TOK_OP_AND:
      return (par_op_ea (par, 0x21, M_ALL));

    case TOK_OP_ASL:
      return (par_op_ea (par, 0x02, M_ZPG | M_DIR | M_ZPG_X | M_DIR_X | M_A));

    case TOK_OP_BCC:
      return (par_op_bra (par, COND_CC));

    case TOK_OP_BCS:
      return (par_op_bra (par, COND_CS));

    case TOK_OP_BEQ:
      return (par_op_bra (par, COND_EQ));

    case TOK_OP_BNE:
      return (par_op_bra (par, COND_NE));

    case TOK_OP_BPL:
      return (par_op_bra (par, COND_PL));

    case TOK_OP_BMI:
      return (par_op_bra (par, COND_MI));

    case TOK_OP_BVC:
      return (par_op_bra (par, COND_VC));

    case TOK_OP_BVS:
      return (par_op_bra (par, COND_VS));

    case TOK_OP_BIT:
      return (par_op_ea (par, 0x20, M_ZPG | M_DIR));

    case TOK_OP_BRK:
      return (par_op_imp (par, 0x00));

    case TOK_OP_CLC:
      return (par_op_imp (par, 0x18));

    case TOK_OP_CLD:
      return (par_op_imp (par, 0xd8));

    case TOK_OP_CLI:
      return (par_op_imp (par, 0x58));

    case TOK_OP_CLV:
      return (par_op_imp (par, 0xb8));

    case TOK_OP_CMP:
      return (par_op_ea (par, 0xc1, M_ALL));

    case TOK_OP_CPX:
      return (par_op_cpx (par, 0xe0));

    case TOK_OP_CPY:
      return (par_op_cpx (par, 0xc0));

    case TOK_OP_DB:
      return (par_op_db (par));

    case TOK_OP_DEC:
      return (par_op_ea (par, 0xc2, M_ZPG | M_DIR | M_ZPG_X | M_DIR_X));

    case TOK_OP_DEX:
      return (par_op_imp (par, 0xca));

    case TOK_OP_DEY:
      return (par_op_imp (par, 0x88));

    case TOK_OP_EOR:
      return (par_op_ea (par, 0x41, M_ALL));

    case TOK_OP_INC:
      return (par_op_ea (par, 0xe2, M_ZPG | M_DIR | M_ZPG_X | M_DIR_X));

    case TOK_OP_INX:
      return (par_op_imp (par, 0xe8));

    case TOK_OP_INY:
      return (par_op_imp (par, 0xc8));

    case TOK_OP_JMP:
      return (par_op_jmp (par));

    case TOK_OP_JSR:
      return (par_op_jsr (par));

    case TOK_OP_LDA:
      return (par_op_ea (par, 0xa1, M_ALL));

    case TOK_OP_LDX:
      return (par_op_ldxy (par, 0xa2, 1));

    case TOK_OP_LDY:
      return (par_op_ldxy (par, 0xa0, 0));

    case TOK_OP_LSR:
      return (par_op_ea (par, 0x42, M_ZPG | M_A | M_DIR | M_ZPG_X | M_DIR_X));

    case TOK_OP_NOP:
      return (par_op_imp (par, 0xea));

    case TOK_OP_ORA:
      return (par_op_ea (par, 0x01, M_ALL));

    case TOK_OP_PHA:
      return (par_op_imp (par, 0x48));

    case TOK_OP_PHP:
      return (par_op_imp (par, 0x08));

    case TOK_OP_PLA:
      return (par_op_imp (par, 0x68));

    case TOK_OP_PLP:
      return (par_op_imp (par, 0x28));

    case TOK_OP_ROL:
      return (par_op_ea (par, 0x22, M_ZPG | M_A | M_DIR | M_ZPG_X | M_DIR_X));

    case TOK_OP_ROR:
      return (par_op_ea (par, 0x62, M_ZPG | M_A | M_DIR | M_ZPG_X | M_DIR_X));

    case TOK_OP_RTI:
      return (par_op_imp (par, 0x40));

    case TOK_OP_RTS:
      return (par_op_imp (par, 0x60));

    case TOK_OP_SBC:
      return (par_op_ea (par, 0xe1, M_ALL));

    case TOK_OP_SEC:
      return (par_op_imp (par, 0x38));

    case TOK_OP_SED:
      return (par_op_imp (par, 0xf8));

    case TOK_OP_SEI:
      return (par_op_imp (par, 0x78));

    case TOK_OP_STA:
      return (par_op_ea (par, 0x81, M_ALL & ~M_IMM));

    case TOK_OP_STX:
      return (par_op_stxy (par, 0x82, 1));

    case TOK_OP_STY:
      return (par_op_stxy (par, 0x80, 0));

    case TOK_OP_TAX:
      return (par_op_imp (par, 0xaa));

    case TOK_OP_TAY:
      return (par_op_imp (par, 0xa8));

    case TOK_OP_TSX:
      return (par_op_imp (par, 0xba));

    case TOK_OP_TXA:
      return (par_op_imp (par, 0x8a));

    case TOK_OP_TXS:
      return (par_op_imp (par, 0x9a));

    case TOK_OP_TYA:
      return (par_op_imp (par, 0x98));

    case TOK_OP_REP:
      return (par_op_rep (par));

    default:
      err_print (par, ERR_ANY, "Unknown mnemonic");
      return (1);
  }

  return (0);
}

int par_org (parser_t *par)
{
  unsigned addr;

  par_scan (par);

  if (par_id (par) == TOK_SYM_PC) {
    addr = (par_get_base (par) + cg_get_index (par->cg)) & 0xffff;
    par_scan (par);
  }
  else {
    if (par_exp (par, &addr)) {
      return (1);
    }
  }

  cg_set_addr (par->cg, addr);

  return (0);
}

int par_get_local_label (parser_t *par, char *dst, char *src)
{
  unsigned i, j;
  char     *str;

  i = 0;
  str = par->label;

  while ((i < (PAR_MAX_LABEL - 1)) && (str[i] != 0)) {
    dst[i] = str[i];
    i += 1;
  }

  if (i < (PAR_MAX_LABEL - 1)) {
    dst[i++] = '.';
  }

  j = 0;

  while ((i < (PAR_MAX_LABEL - 1)) && (src[j] != 0)) {
    dst[i] = src[j];
    i += 1;
    j += 1;
  }

  dst[i] = 0;

  return (0);
}

int par_add_label (parser_t *par, char *str, unsigned val)
{
  symbol_t *sym;

  sym = st_sym_get (par->st, str);

  if (sym != NULL) {
    if (sym->pass == par->pass) {
      err_print (par, ERR_ANY, "Symbol already defined");
      return (1);
    }

    if (val != sym->val) {
      par->newpass = 1;
    }

    sym->val = val;
    sym->pass = par->pass;

    return (0);
  }

  sym = sym_new (str, val);
  if (sym == NULL) {
    return (1);
  }

  sym->pass = par->pass;

  st_sym_add (par->st, sym);

  return (0);
}

int par_local_label (parser_t *par)
{
  unsigned val;
  char     str[PAR_MAX_LABEL];

  par_scan (par);

  if (par_id (par) != TOK_IDENT) {
    err_print (par, ERR_ANY, "Local label expected");
    return (1);
  }

  if (par_get_local_label (par, str, par_str (par))) {
    return (1);
  }

  val = cg_get_addr (par->cg);

  par_scan (par);

  if (par_id (par) == TOK_SYM_COLON) {
    par_scan (par);
  }

  if (par_add_label (par, str, val)) {
    return (1);
  }

  return (0);
}

int par_label (parser_t *par)
{
  unsigned val;
  char     str[PAR_MAX_LABEL];

  val = cg_get_addr (par->cg);
  str_copy (str, par_str (par), PAR_MAX_LABEL);

  par_scan (par);

  if (par_id (par) == TOK_SYM_COLON) {
    par_scan (par);
  }

  if (par_id (par) == TOK_SYM_EQU) {
    par_scan (par);

    if (par_exp (par, &val)) {
      return (1);
    }
  }

  par_set_label (par, str);

  if (par_add_label (par, str, val)) {
    return (1);
  }

  return (0);
}

int par_include (parser_t *par)
{
  scanner_t *scn;

  par_scan (par);

  if (par_id (par) != TOK_STRING) {
    err_print (par, ERR_SYNTAX, "File name expected");
    return (1);
  }

  scn = scn64_new ();

  if (scn64_set_file (scn, par_str (par))) {
    err_print (par, ERR_ANY, "Can't open include file");
    scn->del (scn);
    return (1);
  }

  par_scan (par);

  scn64_set_next (scn, par->scn);
  par->scn = scn;

  par_scan (par);

  return (0);
}

int par_module (parser_t *par)
{
  symbol_t  *sym;

  if (par_id (par) != TOK_SYM_MODULE) {
    err_print (par, ERR_SYNTAX, "'module' expected");
    return (1);
  }

  par_scan (par);
  if (par_id (par) != TOK_IDENT) {
    err_print (par, ERR_SYNTAX, "Module name expected");
    return (1);
  }

  par->mod_name = str_clone (par_str (par));
  if (par_label (par)) {
    return (1);
  }

  sym = st_sym_get (par->st, par->mod_name);
  if (sym == NULL) {
    return (1);
  }

  cg_set_addr (par->cg, sym_get_val (sym));
  par_set_base (par, sym_get_val (sym));

  while (par_id (par) != TOK_SYM_END) {
    if (par_is_mnemonic (par)) {
      if (par_mnemonic (par)) {
        return (1);
      }
    }
    else if (par_id (par) == TOK_IDENT) {
      if (par_label (par)) {
        return (1);
      }
    }
    else if (par_id (par) == TOK_SYM_PERIOD) {
      if (par_local_label (par)) {
        return (1);
      }
    }
    else if (par_id (par) == TOK_SYM_INCLUDE) {
      if (par_include (par)) {
        return (1);
      }
    }
    else if (par_id (par) == TOK_SYM_ORG) {
      if (par_org (par)) {
        return (1);
      }
    }
    else {
      err_print (par, ERR_SYNTAX, "'end' expected");
      return (1);
    }
  }

  return (0);
}
